"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.generatePersistedQueryManifest = exports.defaults = void 0;
const cache_1 = require("@apollo/client/cache");
const core_1 = require("@apollo/client/core");
const persisted_query_lists_1 = require("@apollo/persisted-query-lists");
const graphql_tag_pluck_1 = require("@graphql-tools/graphql-tag-pluck");
const glob_1 = require("glob");
const graphql_1 = require("graphql");
const lodash_1 = require("lodash");
const node_crypto_1 = require("node:crypto");
const node_fs_1 = require("node:fs");
const vfile_1 = __importDefault(require("vfile"));
const vfile_reporter_1 = __importDefault(require("vfile-reporter"));
const chalk_1 = __importDefault(require("chalk"));
exports.defaults = {
    documents: "src/**/*.{graphql,gql,js,jsx,ts,tsx}",
    documentIgnorePatterns: [
        "**/*.d.ts",
        "**/*.spec.{js,jsx,ts,tsx}",
        "**/*.story.{js,jsx,ts,tsx}",
        "**/*.test.{js,jsx,ts,tsx}",
    ],
    output: "persisted-query-manifest.json",
};
const COLORS = {
    filepath: chalk_1.default.underline.cyan,
    name: chalk_1.default.yellow,
};
const ERROR_MESSAGES = {
    anonymousOperation: (node) => {
        return `Anonymous GraphQL operations are not supported. Please name your ${node.operation}.`;
    },
    uniqueFragment: (name, source) => {
        return `Fragment named "${COLORS.name(name)}" already defined in: ${COLORS.filepath(source.file.path)}`;
    },
    uniqueOperation: (name, source) => {
        return `Operation named "${COLORS.name(name)}" already defined in: ${COLORS.filepath(source.file.path)}`;
    },
};
function addError(source, message) {
    const vfileMessage = source.file.message(message, source.location);
    vfileMessage.fatal = true;
}
function getDocumentSources(filepath) {
    const file = (0, vfile_1.default)({
        path: filepath,
        contents: (0, node_fs_1.readFileSync)(filepath, "utf-8"),
    });
    if (file.extname === ".graphql" || file.extname === ".gql") {
        return [
            {
                node: (0, graphql_1.parse)(file.toString()),
                file,
                location: { line: 1, column: 1 },
            },
        ];
    }
    return (0, graphql_tag_pluck_1.gqlPluckFromCodeStringSync)(filepath, file.toString()).map((source) => ({
        node: (0, graphql_1.parse)(source.body),
        file,
        location: source.locationOffset,
    }));
}
async function generatePersistedQueryManifest(config = {}) {
    const paths = config.documents ?? exports.defaults.documents;
    const ignorePaths = config.documentIgnorePatterns ?? exports.defaults.documentIgnorePatterns;
    const filepaths = await (0, glob_1.glob)(paths, { ignore: ignorePaths });
    const sources = [...new Set(filepaths)].flatMap(getDocumentSources);
    const fragmentsByName = new Map();
    const operationsByName = new Map();
    for (const source of sources) {
        (0, graphql_1.visit)(source.node, {
            FragmentDefinition(node) {
                const name = node.name.value;
                const sources = fragmentsByName.get(name) ?? [];
                if (sources.length) {
                    sources.forEach((sibling) => {
                        addError(source, ERROR_MESSAGES.uniqueFragment(name, sibling));
                        addError(sibling, ERROR_MESSAGES.uniqueFragment(name, source));
                    });
                }
                fragmentsByName.set(name, [...sources, source]);
                return false;
            },
            OperationDefinition(node) {
                const name = node.name?.value;
                if (!name) {
                    addError(source, ERROR_MESSAGES.anonymousOperation(node));
                    return false;
                }
                const sources = operationsByName.get(name) ?? [];
                if (sources.length) {
                    sources.forEach((sibling) => {
                        addError(source, ERROR_MESSAGES.uniqueOperation(name, sibling));
                        addError(sibling, ERROR_MESSAGES.uniqueOperation(name, source));
                    });
                }
                operationsByName.set(name, [...sources, source]);
                return false;
            },
        });
    }
    if (sources.some(({ file }) => file.messages.length > 0)) {
        const files = [...new Set(sources.map((source) => source.file))];
        console.error((0, vfile_reporter_1.default)(files, { quiet: true }));
        process.exit(1);
    }
    const fragments = (0, cache_1.createFragmentRegistry)(...sources.map(({ node }) => node));
    const manifestOperations = [];
    const client = new core_1.ApolloClient({
        cache: new core_1.InMemoryCache({
            fragments,
        }),
        link: new core_1.ApolloLink((operation) => {
            const body = (0, graphql_1.print)((0, persisted_query_lists_1.sortTopLevelDefinitions)(operation.query));
            manifestOperations.push({
                id: (0, node_crypto_1.createHash)("sha256").update(body).digest("hex"),
                name: operation.operationName,
                type: operation.query.definitions.find((d) => d.kind === "OperationDefinition").operation,
                body,
            });
            return core_1.Observable.of({ data: null });
        }),
    });
    for (const [_, sources] of (0, lodash_1.sortBy)([...operationsByName.entries()], lodash_1.first)) {
        for (const source of sources) {
            await client.query({ query: source.node, fetchPolicy: "no-cache" });
        }
    }
    return {
        format: "apollo-persisted-query-manifest",
        version: 1,
        operations: manifestOperations,
    };
}
exports.generatePersistedQueryManifest = generatePersistedQueryManifest;
//# sourceMappingURL=index.js.map